//
//  MainViewController.swift
//  qPadLab
//
//  Created by David Arcos Gutierrez on 12/4/18.
//  Copyright © 2018 David Arcos Gutierrez. All rights reserved.
//

import UIKit

class MainViewController: UIViewController {
    
    // Geometric configuration parameters
    var margin = 0.0
    var deltaGrid = 0.0
    var horitzontalCells = 50
    var verticalCells = 0
    
    // Other configuration parameters
    let longPressTime = 0.2
    var interfaceInitialized:Bool = false
    
    // View zones outlets
    @IBOutlet weak var toolBarView: UIView?
    @IBOutlet weak var cellInformationView: UIView?
    @IBOutlet weak var gridContainerView: UIView?
    @IBOutlet weak var gateBoxView: UIView?
    @IBOutlet weak var garbageView: UIView?
    
    // Cell information outlets
    @IBOutlet weak var cellNameLabel: UILabel?
    @IBOutlet weak var fomQuantumCostLabel: UILabel?
    @IBOutlet weak var fomRegisterSizeLabel: UILabel?
    @IBOutlet weak var fomDelayLabel: UILabel?
    @IBOutlet weak var fomGarbageLabel: UILabel?
    
    // Additional outlets
    @IBOutlet weak var icon: UILabel?
    @IBOutlet weak var basket: UILabel?
    
    // Current state variables
    var myTouchPosition = CGPoint()
    var myTouchTime = Date()
    var currentGateTouched: QBlockView?
    var currentGateStatus: QBlockEditableSubstate?
    var currentGateOriginalPosition: CGPoint = CGPoint.zero
    var currentCell: QCell = DataManager.shared.currentCell
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Do any additional setup after loading the view.
        
        // Do initial interface setup and general style improvements
        interfaceSetup()
        
        // Do notifications registration
        registerComunication()
        
        // Figures of merit data initialization
        updateCircuitData()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        
        // First time needs grid initialization (when constraints are applied)
        if interfaceInitialized == false {
            if let v = gridContainerView {
                applyStyleToInitGridContainerView(view: v)
                drawGrid(view: v)
                interfaceInitialized = true
            }
            // Draw current circuit
            loadCircuit(cell: currentCell)
        }
    }
    
    /* ------------------------------------------------- */
    /*                BUTTON ACTIONS                     */
    /* ------------------------------------------------- */
    
    // Deprecated (for test only)
    @IBAction func closeButton (sender: UIButton) {
        //self.dismiss(animated:true, completion:nil)
    }
    
    @IBAction func openExportationView (sender: UIButton) {
        
        let vc = ExportationViewController(
            nibName: "ExportationViewController",
            bundle: nil)
        //vc.currentCell = self.currentCell
        self.present(vc, animated:true, completion:nil)
    }
    
    @IBAction func openConfigView (sender: UIButton) {
        
        let vc = ConfigViewController(
            nibName: "ConfigViewController",
            bundle: nil)
        self.present(vc, animated:true, completion:nil)
    }
    
    @IBAction func openHelpView (sender: UIButton) {
        
        let vc = HelpViewController(
            nibName: "HelpViewController",
            bundle: nil)
        self.present(vc, animated:true, completion:nil)
    }
    
    @IBAction func newCellButton (sender: UIButton) {
        
        if currentCell.unsavedChanges && currentCell.gates.count != 0 {
            self.showUnsavedChangesAlert(){
                self.cleanCurrentCircuit()
            }
        }else{
            self.cleanCurrentCircuit()
        }
    }
    
    @IBAction func loadCellButton (sender: UIButton) {
        
        if currentCell.unsavedChanges && currentCell.gates.count != 0 {
            self.showUnsavedChangesAlert(){
                let vc = LibraryViewController(
                    nibName: "LibraryViewController",
                    bundle: nil)
                self.present(vc, animated:true, completion:nil)
            }
        }else{
            let vc = LibraryViewController(
                nibName: "LibraryViewController",
                bundle: nil)
            self.present(vc, animated:true, completion:nil)
        }
    }
    
    @IBAction func saveCellButton (sender: UIButton) {
        
        if  currentCell.unregisteredCell{
            self.showChangeNameAlert(cleanGrid: false){}
        }else if currentCell.unsavedChanges{
            self.saveCurrentCircuit()
        }
    }
    
    @IBAction func load (sender: UIButton) {
        DataManager.shared.loadCirctuits()
    }
    
    /* ------------------------------------------------- */
    /*                  INTERFACE                        */
    /* ------------------------------------------------- */
    func interfaceSetup (){
        
        if let v = toolBarView {
            applyStyleToInitToolBarView(view:v)
        }
        
        if let v = cellInformationView {
            applyStyleToInitCellInformationView(view: v)
        }
        
        if let v = gateBoxView {
            applyStyleToInitGateBoxView(view: v)
            fillGateBox(view: v)
        }
        
        if let v = garbageView {
            applyStyleToInitGarbageView(view: v)
        }
    }
    
    // Draw grid dots
    func drawGrid(view: UIView) {
        
        self.margin = StyleManager.shared.generalViewPadding
        let dotSize = StyleManager.shared.dotSize
        self.horitzontalCells = StyleManager.shared.horitzontalPoints
        
        // Calculate available space
        let availableHoritzontalSpace = Double(view.frame.size.width) - 2.0*margin
        let availableVerticalSpace = Double(view.frame.size.height) - 2.0*margin
        
        // Initialize grid parameters based on available space
        self.deltaGrid = availableHoritzontalSpace/Double(horitzontalCells)
        self.verticalCells = min(Int((availableVerticalSpace)/deltaGrid), StyleManager.shared.maxVerticalPoints)
        
        // Add grid dots
        for i in 0...horitzontalCells {
            for j in 0...verticalCells {
                let v = UIView()
                v.frame = CGRect(x: margin - dotSize/2.0 + deltaGrid*Double(i),
                                 y: margin - dotSize/2.0 + deltaGrid*Double(j),
                                 width: dotSize,
                                 height: dotSize)
                v.layer.cornerRadius = CGFloat(dotSize/2.0)
                
                v.backgroundColor = UIColor.init(red: 0.5,
                                                 green: 0.5,
                                                 blue: 0.5,
                                                 alpha: 0.2)
                
                view.addSubview(v)
            }
        }
    }
    
    // Fill gatebox with a set of basic gates
    func fillGateBox(view: UIView) {
        
        // Configuration parameters
        let columns = 2
        let padding = 10.0
        
        // Calculate gate size (squared)
        let boxSize = (Double(view.frame.width) - Double(columns+1)*padding)/Double(columns)
        
        // Array of gates
        var gates:[QBlockView] = Array()
        gates.append(QBlockView.init(type: .hadamard))
        gates.append(QBlockView.init(type: .pauliX))
        gates.append(QBlockView.init(type: .pauliY))
        gates.append(QBlockView.init(type: .pauliZ))
        gates.append(QBlockView.init(type: .feynman))
        
        // Create QBlockViews from array and add them as gatebox subviews
        for (index, gate) in gates.enumerated() {
            
            let gateFrame = CGRect(x: padding + Double(index%columns)*(padding + boxSize),
                                   y: padding + Double(index/columns)*(padding + boxSize),
                                   width: boxSize,
                                   height: boxSize)
            let newGate:QBlockView
            
            switch(gate.type) {
                case .hadamard:
                    newGate = HadamardGate.init(frame: gateFrame)
                case .pauliX:
                    newGate = PauliGate.init(frame: gateFrame, type:.pauliX)
                case .pauliY:
                    newGate = PauliGate.init(frame: gateFrame, type:.pauliY)
                case .pauliZ:
                    newGate = PauliGate.init(frame: gateFrame, type:.pauliZ)
                case .feynman:
                    newGate = FeynmanGate.init(frame: gateFrame, inputs: 3, reversed: false)
                default:
                    newGate = HadamardGate.init(frame: gateFrame)
            }
            
            newGate.state = .gatebox
            view.addSubview(newGate)
        }
    }
    
    // Get valid grid coordinates from an arbitrary view origin coordinates
    @discardableResult func gridCoordinatesFromOrigin(cell:QBlockView) -> (x: CGFloat, y: CGFloat){
        
        var gridXOffset = 0.0
        var gridYOffset = 0.0
        
        // Origin translation (set offset)
        if let v = gridContainerView {
            gridXOffset = Double(v.frame.origin.x)
            gridYOffset = Double(v.frame.origin.y)
        }
        
        // View origin coordinates
        let x = Double(cell.frame.origin.x)
        let y = Double(cell.frame.origin.y)
        
        // Calculate grid coordinates
        let column = floor((x - margin - gridXOffset)/deltaGrid)
        let row = floor((y - margin - gridYOffset)/deltaGrid)
        
        // Update gate's model
        cell.gridOrigin.x = Int(column)
        cell.gridOrigin.y = Int(row)
        
        // Calculate modified coordinates (valid grid coordinates)
        let newXCoord = margin + gridXOffset + column*deltaGrid
        let newYCoord = margin + gridYOffset + row*deltaGrid
        
        // Calculate origin correction
        let xCoordinateCorrection = cell.frame.origin.x - CGFloat(newXCoord)
        let yCoordinateCorrection = cell.frame.origin.y - CGFloat(newYCoord)
        
        // Update view's frame
        cell.frame.origin = CGPoint(x:newXCoord, y:newYCoord)
        
        // Return corrections
        return (xCoordinateCorrection, yCoordinateCorrection)
    }
    
    /* ------------------------------------------------- */
    /*                       STYLE                       */
    /* ------------------------------------------------- */
    func applyStyleToInitToolBarView(view: UIView){
        StyleManager.shared.setBottomShadowToView(view: view)
    }
    
    func applyStyleToInitCellInformationView(view: UIView){
        StyleManager.shared.setBottomShadowToView(view: view)
    }
    
    func applyStyleToInitGateBoxView(view: UIView){
        StyleManager.shared.setBottomShadowToView(view: view)
    }
    
    func applyStyleToInitGarbageView(view: UIView){
        
        StyleManager.shared.setBottomShadowToView(view: view)
        
        basket?.text = "\u{e007}"
        basket?.font = UIFont(name:"anthropometricfont",size:90.0)
        basket?.alpha = 0.5
    }
    
    func applyStyleToInitGridContainerView(view: UIView){
        StyleManager.shared.setBottomShadowToView(view: view)
    }
    
    /* ------------------------------------------------- */
    /*                  DATA MANAGEMENT                  */
    /* ------------------------------------------------- */
    
    func updateCircuitData() {
        
        // Show cell state
        if currentCell.unregisteredCell {
            cellNameLabel?.text = "\(currentCell.name) (new)"
        }else if currentCell.gates.count == 0 {
            cellNameLabel?.text = "\(currentCell.name) (empty)"
        }else if currentCell.unsavedChanges {
            cellNameLabel?.text = "\(currentCell.name) (unsaved)"
        }else{
            cellNameLabel?.text = currentCell.name
        }
        
        // Recalculate figures of merit
        recalculateFiguresOfMerit()
    }
    
    func recalculateFiguresOfMerit(){
        
        // Update figures of merit labels
        fomQuantumCostLabel?.text = "Quantum cost: \(CircuitEvaluator.evaluateFigureQuantumCost(cell: currentCell))"
        fomDelayLabel?.text = "Delay: \(CircuitEvaluator.evaluateFigureDelay(cell: currentCell))"
        fomGarbageLabel?.text = "Garbage: \(CircuitEvaluator.evaluateFigureGarbage(cell: currentCell))"
        fomRegisterSizeLabel?.text = "Register size: \(CircuitEvaluator.evaluateFigureRegisterSize(cell: currentCell))"
    }
    
    /* ------------------------------------------------- */
    /*                 INTERACTION                       */
    /* ------------------------------------------------- */
    
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        
        // Save time for long press management
        myTouchTime = Date.init()
        
        let touch:UITouch = touches.first!
        
        /***     C-Not target management     ***/
        if let qTarget = touch.view as? FeynmanTarget {
            
            if let qCnot = qTarget.superview as? FeynmanGate {
                
                // When C-Not's superview state is fixed, then the target is editable
                if qCnot.state.hashValue == QBlockState.fixed.hashValue{
                    StyleManager.shared.setGateBlockStyle(gate: qCnot, style: .highlighted)
                    return
                }
            }
        }
        
        /***     QBlockView management     ***/
        if let qview = touch.view as? QBlockView {
            
            // Save touch position
            myTouchPosition = touch.location(in: self.view)
            
            if qview.state.hashValue == QBlockState.gatebox.hashValue {
            
                let hadamardBoxSize = CGFloat(deltaGrid*4)
                
                let gateFrame = CGRect(x: myTouchPosition.x-hadamardBoxSize/2,
                                       y: myTouchPosition.y-hadamardBoxSize/2,
                                       width: CGFloat(hadamardBoxSize),
                                       height: CGFloat(hadamardBoxSize))
                
                // Create a copy of the selected gate (from gatebox) and set it as the current gate touched
                if let qview = touch.view as? PauliGate {
                    currentGateTouched = PauliGate.init(frame: gateFrame, type:qview.type)
                } else if let qview = touch.view as? FeynmanGate {
                    let feynameBoxSize = CGFloat(deltaGrid*6)
                    let feynmanGateFrame = CGRect(x: myTouchPosition.x-feynameBoxSize/2,
                                                  y: myTouchPosition.y-feynameBoxSize/2,
                                                  width: CGFloat(hadamardBoxSize),
                                                  height: CGFloat(feynameBoxSize))
                    currentGateTouched = FeynmanGate.init(frame: feynmanGateFrame, inputs:5, reversed: qview.reversed)
                } else {
                    currentGateTouched = HadamardGate.init(frame: gateFrame)
                }
                
                StyleManager.shared.setGateBlockStyle(gate: currentGateTouched, style: .highlighted)
                
                self.view.addSubview(currentGateTouched!)
            }else{
                
                // Set current gate touched
                currentGateTouched = qview
                
                // Save original gate position
                currentGateOriginalPosition = qview.frame.origin
                
                StyleManager.shared.setGateBlockStyle(gate: qview, style: .highlighted)
                
                // Change current gate touched state to waiting (for long press time)
                qview.state = QBlockState.waiting
                
                // After "longPressTime" seconds, the state becomes editable (long press done)
                DispatchQueue.main.asyncAfter(deadline: .now() + longPressTime) {
                    switch qview.state {
                        case .waiting:
                            // If gate is waiting, then the new state is "editable"
                            qview.state = QBlockState.editable
                            StyleManager.shared.setGateBlockStyle(gate: qview, style: .editing)
                        
                        default: break
                    }
                }
            }
        }
    }

    override func touchesMoved(_ touches: Set<UITouch>, with event: UIEvent?) {
        
        let touch:UITouch = touches.first!
        
        /***     C-Not target management     ***/
        if let qTarget = touch.view as? FeynmanTarget {
            if let qCnot = qTarget.superview as? FeynmanGate {
                if qCnot.state.hashValue == QBlockState.fixed.hashValue{
                    
                    // Calculate displacement vector from previous position
                    let point = touch.location(in: self.view)
                    let vector = CGPoint(x:point.x - myTouchPosition.x,
                                         y:point.y - myTouchPosition.y)
                    
                    // C-Not target can only be modified through its Y axis
                    if vector.y >= 0 {
                        // When positive displacement is greater than grid separation
                        if vector.y >= CGFloat(deltaGrid) {
                            let hadamardBoxSize = CGFloat(deltaGrid*4)
                            
                            // Increase gate's height
                            qCnot.gridRect.height += 1
                            
                            // Change QBlockView frame
                            let feynameBoxSize = CGFloat(deltaGrid*Double(qCnot.gridRect.height))
                            let feynmanGateFrame = CGRect(x: qCnot.frame.origin.x,
                                                          y: qCnot.frame.origin.y,
                                                          width: CGFloat(hadamardBoxSize),
                                                          height: feynameBoxSize)
                            
                            qCnot.frame = feynmanGateFrame
                            qCnot.setNeedsDisplay()
                            
                            // Update current touch Y coordinate
                            myTouchPosition.y = myTouchPosition.y + vector.y
                        }
                    }else{
                        if qCnot.gridRect.height > 3 {
                            // When negative displacement is greater than grid separation
                            if -vector.y >= CGFloat(deltaGrid) {
                                let hadamardBoxSize = CGFloat(deltaGrid*4)
                                
                                // Decrease gate's height
                                qCnot.gridRect.height -= 1
                                
                                // Change QBlockView frame
                                let feynameBoxSize = CGFloat(deltaGrid*Double(qCnot.gridRect.height))
                                let feynmanGateFrame = CGRect(x: qCnot.frame.origin.x,
                                                              y: qCnot.frame.origin.y,
                                                              width: CGFloat(hadamardBoxSize),
                                                              height: feynameBoxSize)
                                
                                qCnot.frame = feynmanGateFrame
                                qCnot.setNeedsDisplay()
                                
                                // Update current touch Y coordinate
                                myTouchPosition.y = myTouchPosition.y + vector.y
                            }
                        }
                    }
                    return
                }
            }
        }
        
        /***     QBlockView management     ***/
        if let qview = touch.view as? QBlockView {
            if let touch = touches.first {
                
                var myQview = qview
                
                if (qview.state.hashValue == QBlockState.gatebox.hashValue) && currentGateTouched != nil{
                    myQview = currentGateTouched!
                }
                
                // Calculate displacement vector from previous position
                let point = touch.location(in: self.view)
                let vector = CGPoint(x:point.x - myTouchPosition.x,
                                     y:point.y - myTouchPosition.y)
                
                // Long press management. Displacement vector coordinates can't increase before "longPressTime" is completed
                if qview.state.hashValue != QBlockState.gatebox.hashValue {
                    let elapsed = Date().timeIntervalSince(myTouchTime)
                    if elapsed < longPressTime {
                        if vector.x > qview.frame.size.width/2.0 || vector.y > qview.frame.size.height/2.0 {
                            qview.state = QBlockState.fixed //Longpress cancelled (finish waiting)
                        }
                        return
                    }
                }
                
                // If gate state is fixed then it can be modified
                if myQview.state.hashValue == QBlockState.fixed.hashValue {
                    return
                }
                
                // Perform gate's displacement (updating view's center position)
                myQview.center = CGPoint(x: myQview.center.x + vector.x,
                                         y: myQview.center.y + vector.y)
                
                // Reassign touch position
                myTouchPosition = point
                
                if let v = gridContainerView {
                    
                    // If touch is placed inside the grid zone
                    if v.frame.contains(touch.location(in: self.view)) {
                        
                        // Transform gate's position into valid grid position
                        let vectorCorrection = self.gridCoordinatesFromOrigin(cell:myQview)
                        myTouchPosition.x = myTouchPosition.x - vectorCorrection.x
                        myTouchPosition.y = myTouchPosition.y - vectorCorrection.y
                        
                        if (myQview.gridOrigin.x < 0) || (myQview.gridOrigin.y < 0) || (myQview.gridOrigin.x + myQview.gridRect.width > horitzontalCells) || (myQview.gridOrigin.y + myQview.gridRect.height > verticalCells) {
                            // Outside grid zone
                            currentGateStatus = .wrongPlaced
                            StyleManager.shared.setGateBlockStyle(gate: myQview, style: .outsite)
                        }else{
                            // Inside grid zone
                            currentGateStatus = .wellPlaced
                            StyleManager.shared.setGateBlockStyle(gate: myQview, style: .editing)
                        }
                        
                    }else{
                        
                        if let gv = garbageView {
                            
                            // If touch is placed inside the garbage zone
                            if gv.frame.contains(touch.location(in: self.view)){
                                currentGateStatus = .garbagePlaced
                                StyleManager.shared.setGateBlockStyle(gate: myQview, style: .onGarbage)
                            }else{
                                
                                // Then touch is placed outside (invalid zone)
                                currentGateStatus = .wrongPlaced
                                StyleManager.shared.setGateBlockStyle(gate: myQview, style: .outsite)
                            }
                        }
                    }
                }
                
                // Gate intersection detection (with other gates)
                for g in currentCell.gates{
                    if g != myQview {
                        let intersectionRect = g.frame.intersection(myQview.frame)
                        if (intersectionRect.size.width > CGFloat(deltaGrid/2.0)) && (intersectionRect.size.height > CGFloat(deltaGrid/2.0)) {
                            StyleManager.shared.setGateBlockStyle(gate: myQview, style: .wrongPlaced)
                            currentGateStatus = .wrongPlaced
                        }
                    }
                }
            }
        }
        
    }

    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?) {
        
        let touch:UITouch = touches.first!
        
        /***     C-Not target management     ***/
        if let qTarget = touch.view as? FeynmanTarget {
            
            if let qCnot = qTarget.superview as? FeynmanGate {
                
                if qCnot.state.hashValue == QBlockState.fixed.hashValue{
                    // Finish edition
                    StyleManager.shared.setGateBlockStyle(gate: qCnot, style: .normal)
                    return
                }
            }
        }
        
        /***     QBlockView management     ***/
        if let qview = touch.view as? QBlockView, let v = gridContainerView {
            
            // If new gate (from gatebox)
            if qview.state.hashValue == QBlockState.gatebox.hashValue {
                if v.frame.contains(touch.location(in: self.view)) {
                    if currentGateStatus?.hashValue == QBlockEditableSubstate.wellPlaced.hashValue {
                        self.currentCell.addGate(gate: currentGateTouched!)
                        StyleManager.shared.setGateBlockStyle(gate: currentGateTouched, style: .normal)
                        currentGateTouched?.state = QBlockState.fixed
                    }else{
                        //TODO: add delete animation
                        currentGateTouched?.removeFromSuperview()
                    }
                }else{
                    // Outside grid (not placed properly)
                    currentGateTouched?.removeFromSuperview()
                }
            }else if qview.state.hashValue == QBlockState.editable.hashValue {
                
                if currentGateStatus?.hashValue == QBlockEditableSubstate.wellPlaced.hashValue {
                    // Well placed
                    qview.state = QBlockState.fixed
                    self.currentCell.sortGates()
                    StyleManager.shared.setGateBlockStyle(gate: currentGateTouched, style: .normal)
                }else if currentGateStatus?.hashValue == QBlockEditableSubstate.garbagePlaced.hashValue {
                    // Garbage placed (deletion protocol)
                    self.currentCell.removeGate(gate: currentGateTouched!)
                    qview.removeFromSuperview()
                }else{
                    // Wrong placed
                    
                    // Reassign the original position
                    qview.frame.origin = currentGateOriginalPosition
                    gridCoordinatesFromOrigin(cell:qview)

                    qview.state = QBlockState.fixed
                    //TODO: add return to initial position animation
                }
            }else{
                qview.state = .fixed
            }
            
            // Update style and cell's data
            StyleManager.shared.setGateBlockStyle(gate: qview, style: .normal)
            updateCircuitData()
        }
    }
    
    /* ------------------------------------------------- */
    /*                  NOTIFICATIONS                    */
    /* ------------------------------------------------- */
    func registerComunication() {
        
        // Register load circuit notification (sent by LibraryViewController)
        let notif = Notification.Name.init(rawValue: "loadCircuit")
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.loadCircuitNote),
                                               name: notif,
                                               object: nil)
    }
    
    @objc func loadCircuitNote(notification: Notification){
        
        // Call load circuit method if cell exists
        if let circuit = notification.object {
            loadCircuit(cell: circuit as! QCell)
        }
    }
    
    /* ------------------------------------------------- */
    /*               INTERNAL METHODS                    */
    /* ------------------------------------------------- */
    
    // Show unsaved changes alert protocol (depending on cell's state)
    func showUnsavedChangesAlert (completion:@escaping ()->()) {
        
        if currentCell.unregisteredCell {
            
            // First time
            let alert = UIAlertController(title: "Unsaved cell", message: "Do you want to save current cell?", preferredStyle: .alert)
            
            alert.addAction(UIAlertAction(title: "Yes", style: .default, handler: { action in
                // Cell name needed (send escaping block)
                self.showChangeNameAlert(cleanGrid: true, completion: completion)
            }))
            alert.addAction(UIAlertAction(title: "No", style: .cancel, handler: { action in
                // Perform escaping block
                completion()
            }))
            
            self.present(alert, animated: true)
        }else{
            // Second time
            let alert = UIAlertController(title: "Unsaved changes", message: "Do you want to save current changes?", preferredStyle: .alert)
            
            alert.addAction(UIAlertAction(title: "Yes", style: .default, handler: { action in
                self.saveCurrentCircuit()
                // Perform escaping block
                completion()
            }))
            alert.addAction(UIAlertAction(title: "No", style: .cancel, handler: { action in
                // Perform escaping block
                completion()
            }))
            
            self.present(alert, animated: true)
        }
    }
    
    // Show change cell name alert (only called for unregistered cells)
    func showChangeNameAlert (cleanGrid: Bool, completion:@escaping ()->()) {
        
        let alert = UIAlertController(title: "Save cell", message: "Introduce the circuit name", preferredStyle: .alert)
        
        let okOption = UIAlertAction(title: "Ok", style: .default, handler: { [weak alert] (_) in
            let textField = alert?.textFields![0]
            self.currentCell.name = (textField?.text)!
            self.saveCurrentCircuit()
            if cleanGrid {
                self.cleanCurrentCircuit()
            }
            // Perform escaping block
            completion()
            self.updateCircuitData()
        })
        
        alert.addTextField { (textField) in
            textField.text = "New circuit"
            
            // Observe the UITextFieldTextDidChange notification to be notified in the below block when text is changed
            NotificationCenter.default.addObserver(forName: .UITextFieldTextDidChange, object: textField, queue: OperationQueue.main, using:
                {_ in
                    let validTextCount = textField.text?.trimmingCharacters(in: .whitespacesAndNewlines).count ?? 0
                    let textIsNotEmpty = validTextCount > 0
                    
                    // Enable OK Button
                    okOption.isEnabled = textIsNotEmpty
            })
        }
        
        alert.addAction(okOption)
        
        self.present(alert, animated: true)
    }
    
    // Clean current circuit method
    func cleanCurrentCircuit () {
        
        // Remove all gate views from superview
        for gate in self.currentCell.gates {
            gate.removeFromSuperview()
        }
        
        // Clear cell's gates
        self.currentCell.clean()
        
        // Recalculate figures of merit
        self.updateCircuitData()
    }
    
    // Save current circuit data method
    func saveCurrentCircuit() {
        
        DataManager.shared.updateCurrentCellChanges()
        updateCircuitData()
        
        // Show toast
        showToast(message: "Successfully saved")
    }
    
    // Initialize QBlockViews
    func loadCircuit(cell:QCell) {
        
        // Make a copy
        let cellCopy: QCell = cell.copy() as! QCell
        
        // Clear current circuit data
        cleanCurrentCircuit()
        
        var gridXOffset = 0.0
        var gridYOffset = 0.0
        
        // Origin translation (set offset)
        if let v = gridContainerView {
            gridXOffset = Double(v.frame.origin.x)
            gridYOffset = Double(v.frame.origin.y)
        }
        
        // Insert gates' views as subviews
        for gate in cellCopy.gates{
            let gateCopy: QBlockView = gate.copy() as! QBlockView
            
            let gateXCoord = margin + gridXOffset + Double(gate.gridOrigin.x)*deltaGrid
            let gateYCoord = margin + gridYOffset + Double(gate.gridOrigin.y)*deltaGrid
            let gateWidth = deltaGrid*Double(gate.gridRect.width)
            let gateHeight = deltaGrid*Double(gate.gridRect.height)
            
            gateCopy.frame = CGRect(x: CGFloat(gateXCoord),
                                    y: CGFloat(gateYCoord),
                                    width: CGFloat(gateWidth),
                                    height: CGFloat(gateHeight))
            self.view.addSubview(gateCopy)
            currentCell.addGate(gate: gateCopy)
        }
        
        // Update current cell data
        currentCell.name = cellCopy.name
        currentCell.unsavedChanges = cellCopy.unsavedChanges
        currentCell.unregisteredCell = cellCopy.unregisteredCell
        currentCell.identifier = cellCopy.identifier
        updateCircuitData()
    }
    
    // Not used. For test only
    func blockTimeEvaluation() {
        
        // <<<<<<<<<< Start time
        NSLog("Evaluating problem")
        let start = NSDate()
        // <<<<<<<<<< Start time
        
        //Evaluation block
        
        // <<<<<<<<<<   end time
        let end = NSDate()
        let timeInterval: Double = end.timeIntervalSince(start as Date) // <<<<< Difference in seconds (double)
        NSLog("Time to evaluate problem: \(timeInterval) seconds")
        // <<<<<<<<<<   end time
    }
}
