//
//  CodeTranslator.swift
//  qPadLab
//
//  Created by David Arcos Gutierrez on 18/4/18.
//  Copyright © 2018 David Arcos Gutierrez. All rights reserved.
//

import UIKit

public enum QuantumLanguage {
    case QASM       //QASM (IBM Quantum Experience)
    case QScript    //QScript (Quantum Computing Playground)
}

public class CodeTranslator: NSObject {
    
    static let shared = CodeTranslator()
    var preferredLanguage = QuantumLanguage.QASM //Defualt exportation language: QASM
    
    // Copy code string to clipboard
    public func pasteBoardCode(cell: QCell) {
        
        // Use preferred language option
        UIPasteboard.general.string = createCodeString(cell: cell)
    }
    
    public func createCodeString(cell: QCell) -> String {
        
        switch(preferredLanguage) {
        case .QScript:
            return createQScriptString(cell: cell)
        default:
            return createQASMString(cell: cell)
        }
    }
    
    // Create QASM string from cell object data
    public func createQASMString(cell: QCell) -> String {
        
        // Header
        var stringFile = "//Code generated by qPadLab\n"
        stringFile += "OPENQASM 2.0;\n\n"
        stringFile += "include \"qelib1.inc\";\n\n"
        
        // Registers size (both quantum and classical)
        stringFile += "qreg q[\(CircuitEvaluator.evaluateFigureRegisterSize(cell: cell))];\n"
        stringFile += "creg c[\(CircuitEvaluator.evaluateFigureRegisterSize(cell: cell))];\n\n"
        
        // Precondition: gates are correctly sorted before starting translation process
        for gate in cell.gates {
            
            let registerLine = cell.qubitFromCoordinate(y:gate.gridOrigin.y + 2)
            
            switch(gate.type){
                case .feynman:
                    let control = cell.qubitFromCoordinate(y:gate.gridOrigin.y + 1)
                    let target = cell.qubitFromCoordinate(y:gate.gridOrigin.y + gate.gridRect.height - 1)
                    if let feynmanGate = gate as? FeynmanGate {
                        if feynmanGate.reversed{
                            stringFile += "cx q[\(target)],q[\(control)];\n"
                        }else{
                            stringFile += "cx q[\(control)],q[\(target)];\n"
                        }
                    }
                case .hadamard:
                    stringFile += "h q[\(registerLine)];\n"
                case .pauliX:
                    stringFile += "x q[\(registerLine)];\n"
                case .pauliY:
                    stringFile += "y q[\(registerLine)];\n"
                case .pauliZ:
                    stringFile += "z q[\(registerLine)];\n"
                default:
                    stringFile += "//Error on gate translation\n"
            }
        }
        
        return stringFile
    }
    
    // Create QScript string from cell object data
    public func createQScriptString(cell: QCell) -> String {
        
        // Header
        var stringFile = "//Code generated by qPadLab\n"
        stringFile += "//\n"
        
        // Quantum vector size
        let vectorSize = CircuitEvaluator.evaluateFigureRegisterSize(cell: cell)
        stringFile += "VectorSize \(vectorSize)\n\n"
        
        // Precondition: gates are correctly sorted before starting translation process
        for gate in cell.gates {
            
            let registerLine = cell.qubitFromCoordinate(y:gate.gridOrigin.y + 2)
            
            switch(gate.type){
            case .feynman:
                let control = cell.qubitFromCoordinate(y:gate.gridOrigin.y + 1)
                let target = cell.qubitFromCoordinate(y:gate.gridOrigin.y + gate.gridRect.height - 1)
                if let feynmanGate = gate as? FeynmanGate {
                    if feynmanGate.reversed{
                        stringFile += "CNot \(target), \(control)\n"
                    }else{
                        stringFile += "CNot \(control), \(target)\n"
                    }
                }
            case .hadamard:
                stringFile += "Hadamard \(registerLine)\n"
            case .pauliX:
                stringFile += "SigmaX \(registerLine)\n"
            case .pauliY:
                stringFile += "SigmaY \(registerLine)\n"
            case .pauliZ:
                stringFile += "SigmaZ \(registerLine)\n"
            default:
                stringFile += "//Error on gate translation\n"
            }
        }
        
        return stringFile
    }
    
    // Write file data to memory (.txt format)
    public func writeFile(cell: QCell){
        
        let file = "quantumScript.txt"
        
        let text = createCodeString(cell: cell)
        
        if let dir = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first {
            
            let path = dir.appendingPathComponent(file)
            print(path)
            
            // Writing
            do {
                try text.write(to: path, atomically: false, encoding: String.Encoding.utf8)
            }catch {
                NSLog("Error: writing")
            }
            
            // Reading (test)
            do {
                let text2 = try String(contentsOf: path, encoding: String.Encoding.utf8)
                print(text2)
            }catch {
                NSLog("Error: reading")
            }
        }
    }
    
    // Share file data (.txt format) by using UIDocumentInteractionController
    public func shareFile (cell: QCell){
        
        if var topController = UIApplication.shared.keyWindow?.rootViewController {
            while let presentedViewController = topController.presentedViewController {
                topController = presentedViewController
            }
            
            let file = "quantumScript.txt"
            
            let text = createCodeString(cell: cell)
            
            if let dir = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first {
                
                let path = dir.appendingPathComponent(file)
                print(path)
                
                //writing
                do {
                    try text.write(to: path, atomically: false, encoding: String.Encoding.utf8)
                    let docController = UIDocumentInteractionController(url: path)
                    docController.presentOptionsMenu(from: topController.view.frame, in:topController.view, animated:true)
                }catch {
                    NSLog("Error: writing")
                }
            }
        }
    }
}
