//
//  QCell.swift
//  qPadLab
//
//  Created by David Arcos Gutierrez on 19/4/18.
//  Copyright © 2018 David Arcos Gutierrez. All rights reserved.
//

import UIKit

public class QCell: NSObject, NSCoding, NSCopying {
    
    // Management data
    var unsavedChanges:Bool = true
    var unregisteredCell:Bool = true
    
    // Cell data
    var name:String = ""
    var identifier:String = ""
    var registerCodification:[Int: Int] = [Int: Int]() //[Grid Y position][register's qubit number]
    var gates:[QBlockView] = Array()
    
    // This method returns the register's qubit number from a grid Y coordinate.
    // - First non-empty register line is 0
    // - Last non-empty register line is N-1 (register's size - 1)
    public func qubitFromCoordinate(y: Int) -> Int{
        
        if let q = registerCodification[y] {
            return q
        }
        return 0
    }
    
    // Add a gate to the circuit array
    public func addGate(gate: QBlockView) {
        gates.append(gate)
        sortGates()
    }
    
    // Remove a gate from the circuit array
    public func removeGate(gate: QBlockView) {
        if let index = gates.index(of: gate) {
            gates.remove(at: index)
            sortGates()
        }else{
            NSLog("Error: object not found")
        }
    }
    
    // Remove all gates from the circuit array
    public func removeAllGates(){
        gates.removeAll()
        unsavedChanges = true
    }
    
    // Remove all cell's data (it incudes setting a new cell identifier)
    public func clean(){
        identifier = createCellIdentifier()
        removeAllGates()
        unregisteredCell = true
        name = "quantum cell"
    }
    
    // This method sorts the gates by using its grid origin coordinates.
    //  (0,0) first
    //  (0,0) -> (0,1) -> (1,0) -> (1,1)
    public func sortGates(){
        unsavedChanges = true
        gates.sort { ($0.gridOrigin.x, $0.gridOrigin.y) < ($1.gridOrigin.x, $1.gridOrigin.y) }
    }
    
    public init(cellName: String)
    {
        super.init()
        self.name = cellName
        self.identifier = createCellIdentifier()
    }
    
    public func encode(with aCoder: NSCoder) {
        
        // Encode all model attributes to memory
        aCoder.encode(unregisteredCell, forKey: "QCell_unregisteredCell")
        aCoder.encode(unsavedChanges, forKey: "QCell_unsavedChanges")
        aCoder.encode(identifier, forKey: "QCell_identifier")
        aCoder.encode(name, forKey: "QCell_name")
        aCoder.encode(registerCodification, forKey: "QCell_registerCod")
        aCoder.encode(gates, forKey: "QCell_gates")
    }
    
    required public init(coder aDecoder: NSCoder) {
        super.init()
        
        // Decode all model attributes from memory
        unregisteredCell = aDecoder.decodeBool(forKey: "QCell_unregisteredCell")
        unsavedChanges = aDecoder.decodeBool(forKey: "QCell_unsavedChanges")
        identifier = aDecoder.decodeObject(forKey: "QCell_identifier") as! String
        name = aDecoder.decodeObject(forKey: "QCell_name") as! String
        registerCodification = aDecoder.decodeObject(forKey: "QCell_registerCod") as! [Int: Int]
        gates = aDecoder.decodeObject(forKey: "QCell_gates") as! [QBlockView]
    }
    
    override public var description: String {
        return "\(name) => \(gates.count) gates"
    }
    
    // Create a new instance cell with the same data
    public func copy(with zone: NSZone? = nil) -> Any {
        
        let copy = QCell.init(cellName: self.name)
        
        copy.unsavedChanges = self.unsavedChanges
        copy.unregisteredCell = self.unregisteredCell
        copy.registerCodification = self.registerCodification
        copy.identifier = self.identifier
        
        // Create a copy of all the gates of the array
        for gate in self.gates {
            
            NSLog("Original gate: \(gate.type)")
            let myGate = gate.copy() as! QBlockView
            NSLog("Copied gate: \(myGate.type)")
            copy.gates.append(myGate)
        }
        
        return copy
    }
    
    // Create a unique cell identifier (indexation)
    func createCellIdentifier() -> String {
        
        //TODO: improve identifier generation
        return NSUUID().uuidString
    }
}
